#include <libpv_config.h>

#define _XOPEN_SOURCE	/* make sys/ipc.h happy */

#include <errno.h>
#include <getopt.h>
#include <limits.h>	/* CHAR_MAX */
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <libpv.h>
#include <pv_register.h>
#include <pv_set.h>
#include <pv_get.h>
#include <pv_lock.h>
#include <pv_bs.h>
#include <pv_default.h>
#include <pv_jvisu.h>


#define	LIBPV_STORE_VARNAME	"LIBPV_STORE"
#define LIBPV_DESC_VARNAME	"LIBPV_DESC"

#define PV_VAR_MAX		(127)
#define PV_VAL_MAX		(127)

enum pvtool_option {
	OPTION_GET = CHAR_MAX + 1,
	OPTION_SET,
	OPTION_INC,
	OPTION_DEC,
	OPTION_DEFAULT,
	OPTION_DEFAULT_ALL,
	OPTION_INFO,
	OPTION_JVISU,
	OPTION_STORE,
	OPTION_DESC,
	OPTION_HELP,
};

enum pvtool_action {
	ACTION_NOTHING,
	ACTION_GET,
	ACTION_SET,
	ACTION_INC,
	ACTION_DEC,
	ACTION_DEFAULT,
	ACTION_DEFAULT_ALL,
	ACTION_INFO,		/* FIXME: */
	ACTION_JVISU,
};

static int shutdown = 0;

void signalhandler (int sig)
{
	if (sig == SIGPIPE) return;
	if (sig == SIGHUP) {
		fprintf(stderr, "SIGHUP received\n");
	}
	if (sig == SIGTERM || sig == SIGINT) {
		fprintf(stderr, "received SIGTERM or SIGINT, shutting down\n");
		shutdown = 1;
	}
}

static void print_usage(char *prog)
{
	fprintf(stderr,
		"\n"
		"usage: %s cfg-options [ cmd-options ] [ variable ] [ value ]\n"
		"\n"
		"Command Options:\n"
		"\n"
		"    --get variable       Search for PV 'variable' and return it's value.\n"
		"    --set variable=value Set PV 'variable' to 'value' if value can reasonably be\n"
		"                         converted to the type of the PV.\n"
		"    --inc variable       Increment PV 'variable'.\n"
		"    --dec variable       Decrement PV 'variable'.\n"
		"    --default variable   Set PV 'variable' to the default value specified in\n"
		"                         XML description file.\n"
		"    --default-all        Set all PVs to their default values specified\n"
		"                         in XML description file.\n"
/* 		"    --info --type variable   Output information about this variable\n" */
/* 		"           --size            separated by whitespace.\n" */
/* 		"           --default\n" */
/* 		"           --offset\n" */
		"Config Options:\n"
		"\n"
		"    -d, --desc <file>    Path to XML description file. If this option is not\n"
		"                         available, look at %s in environment.\n"
		"    -s, --store <file>   Path to backing store. If this option is not available,\n"
		"                         look at %s in environment.\n"
		"    -8, --8bit           use 8 bit backing store\n"
		"\n"
		"Alternatively:\n"
		"\n"
		"    -j, --jvisu [<port>] Start JVisu server on port <port> (default = 2202)\n"
		"                         and block until ctrl-c is pressed.\n"
		"\n",
		prog, LIBPV_STORE_VARNAME, LIBPV_DESC_VARNAME);
}


static inline int do_set (struct pv_engine	*engine,
			  struct pv		*pv,
			  char			*val_str)
{
	union {
		long int		l;
		long long int		ll;
		unsigned long int	ul;
		unsigned long long int	ull;
		float			f;
		double			d;
	} val;

	switch (pv->type) {
	case PV_TYPE_INT8:
	case PV_TYPE_INT16:
	case PV_TYPE_INT32:
		val.l = strtol(val_str, NULL, 0);
		if (val.l == 0 && val_str[0] != '0')
			return -EINVAL;

		switch (pv->type) {	/* check it */
		case PV_TYPE_INT8:
			if (val.l < INT8_MIN || val.l > INT8_MAX)
				return -ERANGE;
			break;
		case PV_TYPE_INT16:
			if (val.l < INT16_MIN || val.l > INT16_MAX)
				return -ERANGE;
			break;
		case PV_TYPE_INT32:
			if (val.l < INT32_MIN || val.l > INT32_MAX
			    || ((val.l == LONG_MIN || val.l == LONG_MAX) && errno == ERANGE))
				return -ERANGE;
			break;
		default:
			break;
		}
		break;
	case PV_TYPE_UINT8:
	case PV_TYPE_UINT16:
	case PV_TYPE_UINT32:
		val.ul = strtol(val_str, NULL, 0);
		if (val.ul == 0 && val_str[0] != '0')
			return -EINVAL;

		switch (pv->type) {	/* check it */
		case PV_TYPE_UINT8:
			if (val.ul > UINT8_MAX)
				return -ERANGE;
			break;
		case PV_TYPE_UINT16:
			if (val.ul > INT16_MAX)
				return -ERANGE;
			break;
		case PV_TYPE_UINT32:
			if (val.ul > UINT32_MAX
			    || (val.ul == ULONG_MAX && errno == ERANGE))
				return -ERANGE;
			break;
		default:
			break;
		}
		break;
	case PV_TYPE_INT64:
		val.ll = strtoll(val_str, NULL, 0);
		if (val.ll == 0 && val_str[0] != '0')
			return -EINVAL;

		if ((val.ll == LLONG_MIN || val.ll == LLONG_MAX)
		    && errno == ERANGE)
			return -ERANGE;
		break;
	case PV_TYPE_UINT64:
		val.ull = strtoll(val_str, NULL, 0);
		if (val.ull == 0 && val_str[0] != '0')
			return -EINVAL;

		if (val.ull == ULLONG_MAX && errno == ERANGE)
			return -ERANGE;
		break;
	case PV_TYPE_FLOAT:
		val.f = strtof(val_str, NULL);
		if (val.f == 0 && val_str[0] != '0')
			return -EINVAL;

		if ((val.f == 0 || val.f == -HUGE_VALF || val.f == HUGE_VALF) && errno == ERANGE)
			return -ERANGE;
		break;
	case PV_TYPE_DOUBLE:
		val.d = strtod(val_str, NULL);
		if (val.d == 0.0 && val_str[0] != '0')
			return -EINVAL;

		if ((val.d == 0.0 || val.d == -HUGE_VAL || val.d == HUGE_VAL) && errno == ERANGE)
			return -ERANGE;
		break;

	case PV_TYPE_STRING:
		if (strlen(val_str) > pv->size - 1)
		return -E2BIG;

	case PV_TYPE_INT8_ARRAY:
	case PV_TYPE_UINT8_ARRAY:
	case PV_TYPE_INT16_ARRAY:
	case PV_TYPE_UINT16_ARRAY:
	case PV_TYPE_INT32_ARRAY:
	case PV_TYPE_UINT32_ARRAY:
	case PV_TYPE_FLOAT_ARRAY:
	case PV_TYPE_DOUBLE_ARRAY:
		/* FIXME how to convert from a string to an array type */
		break;

	case PV_TYPE_UNDEF:
		break;
	}

	
	pv_lock_lock(engine);
	switch (pv->type) {
	case PV_TYPE_INT8:
		pv_set_int8(engine, pv, val.l);
		break;
	case PV_TYPE_UINT8:
		pv_set_uint8(engine, pv, val.ul);
		break;

	case PV_TYPE_INT16:
		pv_set_int16(engine, pv, val.l);
		break;
	case PV_TYPE_UINT16:
		pv_set_uint16(engine, pv, val.ul);
		break;


	case PV_TYPE_INT32:
		pv_set_int32(engine, pv, val.l);
		break;
	case PV_TYPE_UINT32:
		pv_set_uint32(engine, pv, val.ul);
		break;


	case PV_TYPE_INT64:
		pv_set_int64(engine, pv, val.ll);
		break;
	case PV_TYPE_UINT64:
		pv_set_uint64(engine, pv, val.ull);
		break;

	case PV_TYPE_STRING:
		strncpy(pv_to_abs_addr(engine, pv)->string, val_str, pv->size);
		pv_to_abs_addr(engine, pv)->string[pv->size - 1] = 0;
		break;

	case PV_TYPE_FLOAT:
		break;

	case PV_TYPE_DOUBLE:
		break;

	case PV_TYPE_INT8_ARRAY:
	case PV_TYPE_UINT8_ARRAY:
	case PV_TYPE_INT16_ARRAY:
	case PV_TYPE_UINT16_ARRAY:
	case PV_TYPE_INT32_ARRAY:
	case PV_TYPE_UINT32_ARRAY:
	case PV_TYPE_FLOAT_ARRAY:
	case PV_TYPE_DOUBLE_ARRAY:
		/* FIXME how to convert from a string to an array type */
		break;

	case PV_TYPE_UNDEF:
		break;
	}
	pv_bs_flush_pv(engine, pv);
	pv_lock_unlock(engine);

	return 0;
}


static int pv_action_set(struct pv_engine	*engine,
			 struct pv		*pv,
			 char			*val)
{
	int	err;

	err = do_set(engine, pv, val);
	if (err) {
		fprintf(stderr, "Unable to set \"%s\" to \"%s\", err %d (%s)\n",
			pv->id, val, err, strerror(-err));
		return err;
	}
	
	fprintf(stdout, "Var \"%s\" set to \"%s\"\n", pv->id, val);
	
	return 0;
}


static int pv_action_get(struct pv_engine	*engine,
			 struct pv		*pv)
{
	pv_lock_lock(engine);
	switch (pv->type) {
	case PV_TYPE_INT8:
		fprintf(stdout, "%d\n", pv_get_int8(engine, pv));
		break;
	case PV_TYPE_UINT8:
		fprintf(stdout, "%u\n", pv_get_uint8(engine, pv));
		break;

	case PV_TYPE_INT16:
		fprintf(stdout, "%d\n", pv_get_int16(engine, pv));
		break;
	case PV_TYPE_UINT16:
		fprintf(stdout, "%u\n", pv_get_uint16(engine, pv));
		break;


	case PV_TYPE_INT32:
		fprintf(stdout, "%d\n", pv_get_int32(engine, pv));
		break;
	case PV_TYPE_UINT32:
		fprintf(stdout, "%d\n", pv_get_uint32(engine, pv));
		break;


	case PV_TYPE_INT64:
		fprintf(stdout, "%lld\n", pv_get_int64(engine, pv));
		break;
	case PV_TYPE_UINT64:
		fprintf(stdout, "%llu\n", pv_get_uint64(engine, pv));
		break;

	case PV_TYPE_STRING:
		fprintf(stdout, "%s\n", pv_to_abs_addr(engine, pv)->string);
		break;

	case PV_TYPE_FLOAT:
		fprintf(stdout, "%f\n", pv_get_float(engine, pv));
		break;

	case PV_TYPE_DOUBLE:
		fprintf(stdout, "%lf\n", pv_get_double(engine, pv));
		break;

	case PV_TYPE_INT8_ARRAY:
	case PV_TYPE_UINT8_ARRAY:
	case PV_TYPE_INT16_ARRAY:
	case PV_TYPE_UINT16_ARRAY:
	case PV_TYPE_INT32_ARRAY:
	case PV_TYPE_UINT32_ARRAY:
	case PV_TYPE_FLOAT_ARRAY:
	case PV_TYPE_DOUBLE_ARRAY:
		break;
			
	case PV_TYPE_UNDEF:
		break;

	}
	pv_lock_unlock(engine);

	return 0;
}


static int pv_action_default(struct pv_engine	*engine,
			     struct pv		*pv)
{
	pv_lock_lock(engine);
	pv_default_pv(engine, pv);
	pv_lock_unlock(engine);

	fprintf(stdout, "Var \"%s\" set to default\n", pv->id);

	return 0;	
}


static int pv_action_default_all(struct pv_engine	*engine)
{
	pv_lock_lock(engine);
	pv_default((pv_engine_t *)engine, NULL);
	pv_lock_unlock(engine);

	fprintf(stdout, "All vars set to default\n");

	return 0;	
}


static int pv_action_jvisu(struct pv_engine *engine, int port)
{
	fprintf(stdout, "starting jvisu server, ctrl-c to stop...\n");
	if (pv_jvisu_init(engine, port)) {
		fprintf(stderr, "failed!\n");
		return 1;
	}
	while (!shutdown) sleep(1);
	fprintf(stdout, "stopping jvisu server\n");
	pv_jvisu_destroy(engine);

	return 0;
}


static int pv_action_incdec(struct pv_engine	*engine,
			    struct pv		*pv,
                            int			add)
{
	pv_lock_lock(engine);
	switch (pv->type) {
	case PV_TYPE_INT8:
		pv_set_int8(engine, pv, pv_get_int8(engine, pv)+add);
		break;
	case PV_TYPE_UINT8:
		pv_set_uint8(engine, pv, pv_get_uint8(engine, pv)+add);
		break;

	case PV_TYPE_INT16:
		pv_set_int16(engine, pv, pv_get_int16(engine, pv)+add);
		break;
	case PV_TYPE_UINT16:
		pv_set_uint16(engine, pv, pv_get_uint16(engine, pv)+add);
		break;


	case PV_TYPE_INT32:
		pv_set_int32(engine, pv, pv_get_int32(engine, pv)+add);
		break;
	case PV_TYPE_UINT32:
		pv_set_uint32(engine, pv, pv_get_uint32(engine, pv)+add);
		break;


	case PV_TYPE_INT64:
		pv_set_int64(engine, pv, pv_get_int64(engine, pv)+add);
		break;
	case PV_TYPE_UINT64:
		pv_set_uint64(engine, pv, pv_get_uint64(engine, pv)+add);
		break;

	case PV_TYPE_STRING:
		fprintf(stderr, "undefined action for string variable\n");
		break;
	case PV_TYPE_FLOAT:
		fprintf(stderr,"undefined action for float variable\n");
		break;
	case PV_TYPE_DOUBLE:
		fprintf(stderr,"undefined action for double variable\n");
		break;
	case PV_TYPE_INT8_ARRAY:
		fprintf(stderr,"undefined action for int array variable\n");
		break;
	case PV_TYPE_UINT8_ARRAY:
		fprintf(stderr,"undefined action for uint array variable\n");
		break;
	case PV_TYPE_INT16_ARRAY:
		fprintf(stderr,"undefined action for int16 array variable\n");
		break;
	case PV_TYPE_UINT16_ARRAY:
		fprintf(stderr,"undefined action for uint16 array variable\n");
		break;
	case PV_TYPE_INT32_ARRAY:
		fprintf(stderr,"undefined action for int32 array variable\n");
		break;
	case PV_TYPE_UINT32_ARRAY:
		fprintf(stderr,"undefined action for uint32 array variable\n");
		break;
	case PV_TYPE_FLOAT_ARRAY:
		fprintf(stderr,"undefined action for float array variable\n");
		break;
	case PV_TYPE_DOUBLE_ARRAY:
		fprintf(stderr,"undefined action for double array variable\n");
		break;
	case PV_TYPE_UNDEF:
		fprintf(stderr,"undefined action for undefined variable\n");
		break;
	}

	pv_lock_unlock(engine);

	return 0;
}


static int pv_action_inc(struct pv_engine	*engine,
			 struct pv		*pv)
{
	return pv_action_incdec(engine, pv, 1);
}


static int pv_action_dec(struct pv_engine	*engine,
			 struct pv		*pv)
{
	return pv_action_incdec(engine, pv, -1);
}


int main(int argc, char *argv[])
{
	struct sigaction	signal_action;
	struct pv_engine	engine;
	pv_engine_attr_t	attr;
	struct pv		*pv = NULL;
	enum pvtool_action	action = ACTION_NOTHING;
	char			*optdesc = NULL;
	char			*optstore = NULL;
	char			*optvar = NULL;
	char			*ptr;
	char			var[PV_VAR_MAX+1];
	char			optval[PV_VAL_MAX+1];
	int			optport = 2202;
	int			opt;
	int			lenvar;
	int			err = EXIT_FAILURE;

	/* setup signal handler */
	signal_action.sa_handler = signalhandler;
	sigemptyset (&signal_action.sa_mask);
	signal_action.sa_flags = 0;
	sigaction (SIGTERM, &signal_action, 0);        /* catches kill from the shell   */
	sigaction (SIGINT,  &signal_action, 0);        /* catches Ctrl-C from the shell */
	sigaction (SIGPIPE, &signal_action, 0);        /* catches Broken Pipe exception */
	sigaction (SIGHUP,  &signal_action, 0);        /* catches HUP                   */

	struct option long_options[] = {
		{ "get",	required_argument, 0, OPTION_GET	},
		{ "set",	required_argument, 0, OPTION_SET	},
		{ "inc",	required_argument, 0, OPTION_INC	},
		{ "dec",	required_argument, 0, OPTION_DEC	},
		{ "default",	required_argument, 0, OPTION_DEFAULT	},
		{ "default-all",no_argument,       0, OPTION_DEFAULT_ALL},
		{ "info",	optional_argument, 0, OPTION_INFO	},
		{ "desc",	required_argument, 0, OPTION_DESC	},
		{ "store",	required_argument, 0, OPTION_STORE	},
		{ "8bit",	no_argument,       0, '8'		},
		{ "jvisu",	optional_argument, 0, OPTION_JVISU	},
		{ "help",	no_argument,       0, OPTION_HELP	},
		{ 0,	        0,                 0, 0                 },
	};
	
	pv_engine_attr_init(&attr);

	/* reset getopt (just in case...) */
	optind = 0;
	while ((opt = getopt_long(argc, argv, "8", long_options, NULL)) != -1) {
		switch (opt) {
		case OPTION_GET:
			if (action)
				goto exit_double_action;
			action = ACTION_GET;
			
			optvar = optarg;
			break;
		case OPTION_SET:
			if (action)
				goto exit_double_action;
			action = ACTION_SET;
			ptr = strchr(optarg, '=');
			if (!ptr || strlen(ptr + 1) == 0) {
				fprintf(stderr, "\na variable must be set in form \"<var>=<val>\"\n");
				goto exit;
			}
			lenvar = ptr - optarg;
			if (lenvar > PV_VAR_MAX) {
				fprintf(stderr, "\nvarname \"%s\" too long\n", optarg);
				goto exit;
			}
			optvar = var;
			strncpy(optvar, optarg, lenvar);
			optvar[lenvar] = 0;

			strncpy(optval, ptr + 1, PV_VAL_MAX);
			optval[PV_VAL_MAX] = 0;
			
			break;
		case OPTION_INC:
			if (action)
				goto exit_double_action;
			action = ACTION_INC;
			optvar = optarg;
			break;
		case OPTION_DEC:
			if (action)
				goto exit_double_action;
			action = ACTION_DEC;
			optvar = optarg;
			break;
		case OPTION_DEFAULT:
			if (action)
				goto exit_double_action;
			action = ACTION_DEFAULT;
			optvar = optarg;
			break;
		case OPTION_DEFAULT_ALL:
			if (action)
				goto exit_double_action;
			action = ACTION_DEFAULT_ALL;
			break;
		case OPTION_INFO:
			if (action)
				goto exit_double_action;
			action = ACTION_INFO;
			if (optvar)
				optvar = optarg;
			break;
		case OPTION_JVISU:
			if (action)
				goto exit_double_action;
			action = ACTION_JVISU;
			if (optarg)
				optport = strtoul(optarg, NULL, 0);
			break;
		case OPTION_DESC:
			optdesc = optarg;
			break;
		case OPTION_STORE:
			optstore = optarg;
			break;
		case '8':
			attr.store_type = PV_STORE_8BIT;
			break;
		case OPTION_HELP:
			err = EXIT_SUCCESS;
			goto exit_usage;
			break;
		default:
			goto exit;
			break;
		}
	}

	if (action == ACTION_NOTHING)
		goto exit_usage;

	if (!optstore)
		optstore = getenv(LIBPV_STORE_VARNAME);
	if (!optdesc)
		optdesc  = getenv(LIBPV_DESC_VARNAME);

	if (!optstore || !optdesc)
		goto exit_usage;

	strncpy(attr.desc_filename,  optdesc,  PV_PATH_MAX);
	strncpy(attr.store_filename, optstore, PV_PATH_MAX);
	
	err = pv_init((pv_engine_t *)&engine, &attr);
	if (err) {
		fprintf(stderr, "failed to init pv_engine, err %d\n", err);
		err = EXIT_FAILURE;
		goto exit;
	}

	if (optvar) {
		pv = pv_register_lookup(&engine, optvar);
		if (!pv) {
			fprintf(stderr, "Unable to find var \"%s\"\n", optvar);
			err = EXIT_FAILURE;
			goto exit_pv_destroy;
		}
	}

	switch (action) {
	case ACTION_GET:
		pv_action_get(&engine, pv);
		break;
	case ACTION_SET:
		pv_action_set(&engine, pv, optval);
		break;
	case ACTION_DEFAULT:
		pv_action_default(&engine, pv);
		break;
	case ACTION_DEFAULT_ALL:
		pv_action_default_all(&engine);
		break;
	case ACTION_JVISU:
		pv_action_jvisu(&engine, optport);
		break;
	case ACTION_INC:
		pv_action_inc(&engine, pv);
		break;
	case ACTION_DEC:
		pv_action_dec(&engine, pv);
		break;
	default:
		fprintf(stderr, "This feature is not implemented yet, sorry. (patches are welcome!)\n");
		break;
	}

	err = EXIT_SUCCESS;

 exit_pv_destroy:
	pv_destroy((pv_engine_t *)&engine);
 exit:
	exit(err);

 exit_double_action:
	fprintf(stderr,
		"\n"
		"********************************\n"
		"*** Only one command, please ***\n"
		"********************************\n");
 exit_usage:	
	print_usage(argv[0]);
	goto exit;
}

